from __future__ import division, print_function


class World(object):
    def __init__(self):
        self.tags = {}
        self.entities = []
        self.aspects = []
        self.modified = set()

    def get_entity(self, tag):
        return self.tags.get(tag)

    def flag_modified(self, entity):
        self.modified.add(entity)

    def add_entity(self, entity):
        entity.set_world(self)
        self.entities.append(entity)

    def remove_entity(self, entity):
        entity.die()

    def add_aspect(self, aspect):
        self.aspects.append(aspect)

    def _delete_entity(self, entity):
        self.entities.remove(entity)
        if entity.tag:
            self.tags.pop(entity.tag)
        entity.unset_world()

    def update(self):
        for entity in self.modified:
            for aspect in self.aspects:
                if aspect.wants(entity) and entity.active:
                    aspect.add(entity)
                else:
                    aspect.discard(entity)
            if not entity.alive:
                self._delete_entity(entity)
        self.modified.clear()


class Entity(object):
    def __init__(self, tag = None, **comps):
        self.world = None
        self.tag = tag
        self.comps = comps
        self.alive = True
        self.active = False

    def set_world(self, world):
        self.world = world
        self.activate()

    def unset_world(self):
        self.world = None
        self.active = False

    def changed(self):
        if self.world:
            self.world.flag_modified(self)

    def __setitem__(self, component_type, component):
        self.comps[component_type] = component
        self.changed()

    def __getitem__(self, component_type):
        return self.comps[component_type]

    def remove(self, component_type):
        self.changed()
        return self.comps.pop(component_type)

    def has(self, component_type):
        return self.comps.has_key(component_type)

    def die(self):
        self.alive = False
        self.deactivate()

    def activate(self):
        self.active = True
        self.changed()

    def deactivate(self):
        self.active = False
        self.changed()


class Aspect(set):
    def __init__(self, world, *component_types):
        self.world = world
        self.comps = component_types
        world.add_aspect(self)

    def wants(self, entity):
        for c in self.comps:
            if not entity.has(c):
                return False
        return True
