from __future__ import division, print_function

import sys

import cards
import es
import events
import factory
import gui
import levels
import render
import settings
import systems
import tools


class State(object):
    def __init__(self, manager, screen):
        self.manager = manager
        self.screen = screen
        self.delta = 1/60
        self.world = es.World()

    def clear_screen(self, colour=(0,0,0)):
        self.screen.fill(colour)

    def render(self):
        self.clear_screen()

    def next_splash_func(self, next_splash):
        return lambda: self.show_next_splash(next_splash)

    def show_next_splash(self, splash):
        self.world.add_entity(splash)

    def splash_sequence(self, images, end_func):
        splashes = [factory.help_splash(image) for image in images]
        tickets = [events.Event("ticket", entity["death ticket"])
                   for entity in splashes]
        for i in xrange(len(splashes)-1):
            events.schedule(tickets[i],
                            self.next_splash_func(splashes[i+1]))
        events.schedule(tickets[-1], end_func)
        self.world.add_entity(splashes[0])


class GameState(State):
    def __init__(self, manager, screen, rules, level=None):
        super(GameState, self).__init__(manager, screen)
        self.rules = rules
        self.level = None if level is None else levels.LEVELS[level]
        self.deck = cards.get_deck(rules.deck)
        self.players = [Player("Dr. Negotium", side="left"),
                        Player("Cpt. Armament", side="right", ai=self.rules.vs_cpu)]
        self.paused = True
        self.initialize()

        self.logic_systems = [
            systems.AiSystem(self),
            systems.ControlSystem(self),
            systems.BattleSystem(self),
            ]

        self.systems = [
            systems.ExpirySystem(self),
            systems.DismissSystem(self),
            render.RenderSystem(self),
            ]

        self.world.add_entity(factory.background("background4.png"))
        self.world.add_entity(factory.sticker(
            "docket.png", settings.DOCKET_POS, layer="mg"))
        self.world.add_entity(factory.sticker(
            "villain.png", settings.VILLAIN_POS, layer="mg"))
        self.world.add_entity(factory.sticker(
            "hero.png", settings.HERO_POS, layer="mg"))

    def update_player_gui(self):
        for p in self.players:
            for each in p.gui:
                each.die()
            p.gui = factory.status_panel(p)
            for each in p.gui:
                self.world.add_entity(each)

    def set_player_status(self, player, status):
        player.status = status
        self.update_player_gui()

    def initialize(self):
        if not self.level:
            self.begin_game()
            return
        self.splash_sequence(self.level.helpimgs, self.begin_game)

    def begin_game(self):
        self.paused = False
        if self.level:
            leveltext = "Stage {}: {}".format(
                self.level.number, self.level.name)
        else:
            leveltext = "{} vs {}".format(self.players[0].name,
                                          self.players[1].name)
        label = factory.label((512,30), leveltext, size=40)
        label["lifespan"] = 4.0
        self.world.add_entity(label)
        self.deal_hand(self.players[0])
        self.deal_hand(self.players[1])
        for p in self.players:
            p.create_control_guides(self.world)
        self.update_player_gui()

    def deal_hand(self, player):
        for entity in player.hand:
            self.deck.release_card(entity["card"])
            entity.die()
        player.hand = [factory.card(self.deck.draw_card(), (0,0))
                       for i in xrange(3)]
        for entity in player.hand:
            self.world.add_entity(entity)
        player.rejigger_hand()

    def deal_card(self, player):
        card = factory.card(self.deck.draw_card(), (0,0))
        self.world.add_entity(card)
        player.hand.append(card)
        player.rejigger_hand()

    def render(self):
        self.clear_screen()
        self.world.update()
        if not self.paused:
            for system in self.logic_systems:
                system.run()
        for system in self.systems:
            system.run()
        events.check_schedule()

    def alert(self, text, depth=100, colour=(255,255,255)):
        banner = factory.banner(text, depth, 1.0, colour=colour)
        self.world.add_entity(banner)

    def finish_level(self):
        if not self.level:
            victor = None
            for player in self.players:
                if player.points >= 3:
                    victor = player.side
            self.manager.change_state(
                VictoryState(self.manager, self.screen, outcome=victor))
            return
        if self.level.number < len(levels.LEVELS):
            next_level = levels.LEVELS[self.level.number]
            self.rules.deck = next_level.deck
            self.manager.change_state(
                GameState(self.manager, self.screen,
                          self.rules, self.level.number))
        else:
            self.manager.change_state(
                VictoryState(self.manager, self.screen))

    def reset_level(self):
        self.manager.change_state(
            GameState(self.manager, self.screen, self.rules, self.level.number-1))

class Player(object):
    hand_position = (80,390)
    card_offset = (128,20)
    
    def __init__(self, name, side, ai=False):
        self.name = name
        self.ai = ai
        self.hand = []
        self.chosen_card = None
        self.side = side
        self.points = 0
        self.status = settings.CHOOSE_STATUS
        self.chance = False
        self.gui = []

    def rejigger_hand(self):
        x,y = self.hand_position
        dx,dy = self.card_offset
        if self.side == "right":
            x = settings.WINDOW_WIDTH - x
            dx *= -1
        for card in self.hand:
            card["rect"].center = (x,y)
            x += dx
            y += dy

    def create_control_guides(self, world):
        if self.ai:
            return
        x,y = self.hand_position
        dx,dy = self.card_offset
        if self.side == "right":
            x = settings.WINDOW_WIDTH - x
            dx *= -1
        keys = "asd" if self.side == "left" else "lkj"
        for i,c in enumerate(keys):
            sticker = factory.sticker("key-{}.png".format(c),
                                      (x + dx*i, 540))
            world.add_entity(sticker)


class MenuState(State):
    def __init__(self, manager, screen):
        super(MenuState, self).__init__(manager, screen)
        self.systems = [
            systems.ExpirySystem(self),
            systems.DismissSystem(self),
            gui.GuiSystem(self),
            render.RenderSystem(self)
            ]
        self.gui = gui.GuiManager(self.world)
        self.initialize()

    def initialize(self):
        title = factory.label((settings.WINDOW_WIDTH//2, 100),
                              "Rock-Paper-Everything", size=100)
        self.world.add_entity(title)
        self.world.add_entity(factory.background("background2.png"))
        self.world.add_entity(factory.label(
            (settings.WINDOW_WIDTH//2, 400),
            "Use the arrow keys and the Enter key in menus.", size=24))
        self.world.add_entity(factory.label(
            (settings.WINDOW_WIDTH//2 + 300, 520),
            "Character art by Lianne Harris :)", size=24))
        self.main_menu()

    def main_menu(self):
        self.gui.set_menu(
            (settings.WINDOW_WIDTH//2, 200),
            [("Single Player", self.game_func(False,True,"primary",0)),
             ("Two Player", lambda: self.deck_menu(False)),
             ("VS Computer", lambda: self.deck_menu(True)),
             ("Quit Game", self.quit_game)]
            )

    def deck_menu(self, cpu):
        options = [(name, self.game_func(True,cpu,deck,None))
                   for name,deck in cards.DECKORDER]
        options.append(("Back", self.main_menu))
        self.gui.set_menu((settings.WINDOW_WIDTH//2, 200),
                          options, title="Select Deck")

    def render(self):
        self.clear_screen()
        self.world.update()
        for system in self.systems:
            system.run()
        events.check_schedule()

    def game_func(self, vs_match, cpu, deck, level):
        rules = Rules(vs_match=vs_match, vs_cpu=cpu, deck=deck)
        return lambda: self.start_game(rules, level)

    def start_game(self, rules, level):
        tools.play_music("grieg.ogg")
        self.manager.change_state(
            GameState(self.manager, self.screen, rules, level))
        

    def quit_game(self):
        print("Thanks for playing!")
        sys.exit()


class Rules(object):
    def __init__(self, **rules):
        self.use_chance = True
        self.vs_cpu = True
        self.vs_match = False
        self.multiplayer = False
        self.deck = None
        self.__dict__.update(rules)


class VictoryState(State):
    def __init__(self, manager, screen, outcome="story"):
        super(VictoryState, self).__init__(manager, screen)
        self.manager = manager
        self.screen = screen
        self.outcome = outcome
        self.end_scene = ["story-10.png",
                          "story-11.png",
                          "story-12.png",
                          "story-13.png",
                          "gratz.png"]
        self.systems = [
            systems.ExpirySystem(self),
            systems.DismissSystem(self),
            render.RenderSystem(self),
            ]
        self.initialize()

    def initialize(self):
        self.world.add_entity(factory.background("background4.png"))
        self.world.add_entity(factory.sticker(
            "docket.png", settings.DOCKET_POS, layer="mg"))
        self.world.add_entity(factory.sticker(
            "villain.png", settings.VILLAIN_POS, layer="mg"))
        self.world.add_entity(factory.sticker(
            "hero.png", settings.HERO_POS, layer="mg"))
        
        if self.outcome == "left":
            self.splash_sequence(["villain-win.png"], self.return_to_menu)
        elif self.outcome == "right":
            self.splash_sequence(["hero-win.png"], self.return_to_menu)
        else:
            self.splash_sequence(self.end_scene, self.return_to_menu)

    def return_to_menu(self):
        tools.fade_music()
        self.manager.change_state(MenuState(self.manager, self.screen))

    def render(self):
        self.clear_screen()
        self.world.update()
        for system in self.systems:
            system.run()
        events.check_schedule()
        
